from typing import List, Tuple, Union, Any

import torch
import torch.nn as nn

from iatransfer.toolkit.base_transfer import Transfer


class MagnitudeTransfer(Transfer):
    """Magnitude tensor transfer algorithm for IAT.
        :param reverse_priority: if set to True, choose channels with smallest weights' sum
        :param input_constrained_by_output: if set to True, choose input channels considering only already chosen output channels. If set to False, choose input channels taking into account all weights.
    """

    def __init__(self, reverse_priority=False,
                 input_constrained_by_output=True, **kwargs) -> None:
        if reverse_priority:
            self.sgn = +1
        else:
            self.sgn = -1
        self.cut_from_tensor = input_constrained_by_output
        super().__init__(**kwargs)

    def update_slice(self, from_tensor: torch.Tensor, to_tensor: torch.Tensor, idx: int,
                     from_slices: List[Union[List[int], slice]], to_slices: List[Union[List[int], slice]]) -> None:
        dims = [i for i in range(len(from_tensor.shape))]
        dims.remove(idx)
        if from_tensor.shape[idx] > to_tensor.shape[idx]:
            if len(dims) == 0:
                out_channels = list(enumerate(from_tensor))
            else:
                out_channels = list(enumerate(from_tensor.abs().sum(dim=dims)))
            out_channels.sort(key=lambda w: self.sgn * w[1])
            c = to_tensor.shape[idx]
            from_ids = sorted([w[0] for w in out_channels[:c]])
            from_slices.append(from_ids)
            to_slices.append(slice(0, c))
            if self.cut_from_tensor:
                from_tensor = torch.index_select(from_tensor, idx, torch.tensor(from_ids))
        else:
            from_slices.append(slice(0, from_tensor.shape[idx]))
            to_slices.append(slice(0, from_tensor.shape[idx]))

    def get_slices(self, from_tensor: torch.Tensor, to_tensor: torch.Tensor) \
            -> Tuple[Tuple[slice, ...], Tuple[Union[Union[slice, torch.Tensor], Any], ...]]:
        if from_tensor is None or to_tensor is None:
            raise ValueError()
        n = len(from_tensor.shape)
        from_slices, to_slices = [], []
        self.update_slice(from_tensor, to_tensor, 0, from_slices, to_slices)
        if n > 1:
            self.update_slice(from_tensor, to_tensor, 1, from_slices, to_slices)

        for a, b in zip(from_tensor.shape[2:], to_tensor.shape[2:]):
            if a < b:
                from_slices.append(slice(0, a))
                to_slices.append(slice((b - a) // 2, -((b - a + 1) // 2)))
            elif a > b:
                from_slices.append(slice((a - b) // 2, -((a - b + 1) // 2)))
                to_slices.append(slice(0, b))
            else:
                from_slices.append(slice(0, a))
                to_slices.append(slice(0, b))
        total_unsqueeze = 0
        for i in range(len(from_slices) - 1, -1, -1):
            if isinstance(from_slices[i], list):
                from_slices[i] = torch.tensor(from_slices[i])
                for _ in range(total_unsqueeze):
                    from_slices[i] = from_slices[i].unsqueeze(-1)
                total_unsqueeze += 1
        return tuple(to_slices), tuple(from_slices)

    def transfer_layer(self, tensor_from: nn.Module, tensor_to: nn.Module, *args, **kwargs) -> None:
        to_slices, from_slices = self.get_slices(tensor_from.weight, tensor_to.weight)
        if tensor_to.weight is not None and tensor_from.weight is not None:
            tensor_to.weight[to_slices] = tensor_from.weight[from_slices]
        if tensor_to.bias is not None and tensor_from.bias is not None:
            to_ids = to_slices[0]
            from_ids = from_slices[0]
            if isinstance(to_ids, torch.Tensor):
                to_ids = to_ids.flatten()
            if isinstance(from_ids, torch.Tensor):
                from_ids = from_ids.flatten()
            tensor_to.bias[to_ids] = tensor_from.bias[from_ids]

    def transfer(self, matched_tensors: List[Tuple[nn.Module, nn.Module]], *args, **kwargs) -> None:
        with torch.no_grad():
            for matching in matched_tensors:
                if isinstance(matching, list):
                    self.transfer(matching)
                elif matching[0] is not None and matching[1] is not None:
                    self.transfer_layer(matching[0], matching[1])



from typing import Tuple, List, Any, Union

import numpy as np
import torch.nn as nn

from iatransfer.toolkit.base_matching import Matching


class DPMatching(Matching):
    """Dynamic programming matching algorithm for IAT.
    """

    def match(self, from_module: List[Union[nn.Module, List[nn.Module]]],
              to_module: List[Union[nn.Module, List[nn.Module]]], *args, **kwargs) \
            -> List[Union[Tuple[nn.Module, nn.Module], List[Tuple[nn.Module, nn.Module]]]]:
        _, matched, _ = self._match_models(from_module, to_module)
        return matched

    def sim(self, from_module: List[Union[nn.Module, List[nn.Module]]],
            to_module: List[Union[nn.Module, List[nn.Module]]]) -> float:
        score, _, _ = self._match_models(from_module, to_module)
        return score / self._match_models(to_module, to_module)[0]

    def _compute_score(self, from_module: nn.Module, to_module: nn.Module) -> float:
        def are_all_of_this_class(layers: List[nn.Module], clazz: Any) -> bool:
            return all([isinstance(layer, clazz) for layer in layers])

        score = 0
        layers = [from_module, to_module]
        if are_all_of_this_class(layers, list):
            score, _, _ = self._match_models(from_module, to_module)
        else:
            classes = [
                nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
                nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
                nn.Linear
            ]
            for clazz in classes:
                if are_all_of_this_class(layers, clazz):
                    score = self.score(from_module, to_module)
                    break

        return score

    def _match_models(self, flat_from_module: Union[nn.Module, List[nn.Module]],
                      flat_to_module: Union[nn.Module, List[nn.Module]]) \
            -> Tuple[float, List[Tuple[nn.Module, nn.Module]], List[Tuple[int, int]]]:
        m = len(flat_from_module)
        n = len(flat_to_module)
        dp = np.zeros((n + 1, m + 1))
        transition = np.zeros((n + 1, m + 1, 2))  #
        scores = np.zeros((n + 1, m + 1))
        # reduction_coeff = 0.7
        for i in range(1, n + 1):
            for j in range(1, m + 1):
                scores[i, j] = self._compute_score(flat_from_module[j - 1], flat_to_module[i - 1])

        for i in range(1, n + 1):
            for j in range(1, m + 1):
                if dp[i - 1, j] > dp[i, j - 1]:
                    dp[i, j] = dp[i - 1, j]
                    transition[i, j] = [i - 1, j]
                else:
                    dp[i, j] = dp[i, j - 1]
                    transition[i, j] = [i, j - 1]
                cumulative_sum = 0
                for k in range(i, 0, -1):
                    cumulative_sum += scores[k, j]
                    # score = cumulative_sum*current_reduction+dp[k-1,j-1]
                    score = cumulative_sum / (i - k + 1) ** 0.5 + dp[k - 1, j - 1]
                    # let d be the number of layers matched in flat_to_module to the current layer in flat_from_module
                    # max possible score = d*f(d)
                    # we want d*f(d) to be increasing - adding more matchings should give better score
                    # we want f(d) to be decreasing - adding more matchings should give lower score per layer,
                    # thanks to it we encourage dynamic programming not to choose single layer all the time
                    if score > dp[i, j]:
                        dp[i, j] = score
                        transition[i, j] = [k - 1, j - 1]
                    # current_reduction*=reduction_coeff

        matched = []
        matched_indices = []
        i, j = n, m

        while i > 0:
            if transition[i, j, 1] == j:
                i -= 1
                matched.append((None, flat_to_module[i]))
                matched_indices.append((None, i))
                continue
            t = transition[i, j, 0]
            j -= 1
            from_model_layer_included = False
            while t < i:
                i -= 1
                if scores[i + 1, j + 1] > 0:
                    if isinstance(flat_from_module[j], list) and isinstance(flat_to_module[i], list):
                        _, sub_matched, sub_matched_indices = self._match_models(flat_from_module[j], flat_to_module[i])
                        matched.append(sub_matched)
                        matched_indices.append(sub_matched_indices)
                        matched_indices.append((j, i))
                    else:
                        matched.append((flat_from_module[j], flat_to_module[i]))
                        matched_indices.append((j, i))
                    from_model_layer_included = True
                else:
                    matched.append((None, flat_to_module[i]))
                    matched_indices.append((None, i))
            if not from_model_layer_included:
                matched_indices.append((j, None))
                matched.append((flat_from_module[j], None))
        matched.reverse()
        matched_indices.reverse()
        return dp[n][m], matched, matched_indices



import torch

from iatransfer.toolkit.base_transfer import Transfer


class FullTransfer(Transfer):
    """Full tensor transfer algorithm for IAT.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def transfer_layer(self, tensor_from: torch.Tensor, tensor_to: torch.Tensor, *args, **kwargs) -> None:
        if tensor_from is None or tensor_to is None:
            return
        from_slices, to_slices = [], []
        for a, b in zip(tensor_from.shape, tensor_to.shape):
            if a < b:
                ids = torch.randint(a, (b,))
                ids[slice((b - a) // 2, -((b - a + 1) // 2))] = torch.arange(a)
                from_slices.append(ids)
            elif a > b:
                from_slices.append((a - b) // 2 + torch.arange(b))
            else:
                from_slices.append(torch.arange(a))
            to_slices.append(slice(0, b))

        total_unsqueeze = 0
        for i in range(len(from_slices) - 1, -1, -1):
            if isinstance(from_slices[i], torch.Tensor):
                for _ in range(total_unsqueeze):
                    from_slices[i] = from_slices[i].unsqueeze(-1)
                total_unsqueeze += 1

        tensor_to[tuple(to_slices)] = tensor_from[tuple(from_slices)]



import random
from typing import Tuple, List, Union

import torch.nn as nn

from iatransfer.toolkit.base_matching import Matching


class RandomMatching(Matching):
    """Random matching algorithm for IAT. Used for baseline.
    """

    def match(self, from_module: List[Union[nn.Module, List[nn.Module]]],
              to_module: List[Union[nn.Module, List[nn.Module]]], *args, **kwargs) \
            -> List[Union[Tuple[nn.Module, nn.Module], List[Tuple[nn.Module, nn.Module]]]]:
        matched = self._match_models(from_module, to_module)
        return matched

    def sim(self, from_module: List[Union[nn.Module, List[nn.Module]]],
            to_module: List[Union[nn.Module, List[nn.Module]]]) -> float:
        return random.random()
    
    def _get_layers_buckets(self, module: nn.Module):
        classes = [
            nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
            nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
            nn.Linear
        ]
        buckets = [[] for _ in range(len(classes))]
        for layer in module:
            for i, clazz in enumerate(classes):
                if isinstance(layer, clazz):
                    buckets[i].append(layer)
        for i in range(len(buckets)):
            random.shuffle(buckets[i])
        return buckets

    def _match_models(self, flat_from_module: List[nn.Module],
                      flat_to_module: List[nn.Module]) \
            -> List[Tuple[nn.Module, nn.Module]]:
        teacher_buckets = self._get_layers_buckets(flat_from_module)
        student_buckets = self._get_layers_buckets(flat_to_module)

        matched = []
        for teacher_bucket, student_bucket in zip(teacher_buckets, student_buckets):
            for teacher_layer, student_layer in zip(teacher_bucket, student_bucket):
                matched.append((teacher_layer, student_layer))
        return matched


from abc import ABC, abstractmethod

import torch.nn as nn


class Score(ABC):
    """Base-class for any scoring algorithm detectable from IAT.
    """

    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs

    @abstractmethod
    def score(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> float:
        """Calculates the float score between two layers.
        """
        pass

    def precompute_scores(self, from_model: nn.Module, to_model: nn.Module, *args, **kwargs) \
            -> float:
        pass

    def __call__(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> float:
        """Alias for 'score'.
        """
        return self.score(from_module, to_module, *args, **kwargs)



from typing import Dict

from torch import nn

from iatransfer.toolkit.base_score import Score
from iatransfer.toolkit.tli import score_autoencoder
from iatransfer.toolkit.tli_helpers import get_model_graph_and_ids_mapping


class AutoEncoderScore(Score):
    """Auto-encoder score algorithm for IAT.
    """

    def precompute_scores(self, from_model: nn.Module, to_model: nn.Module, *args, **kwargs) \
            -> float:
        teacher_graph, teacher_mapping = get_model_graph_and_ids_mapping(from_model)
        student_graph, student_mapping = get_model_graph_and_ids_mapping(to_model)
        self.scores, teacher_arr, student_arr = score_autoencoder(teacher_graph, student_graph)
        self.teacher_mapping = self._join_dicts(teacher_mapping, dict([(x, i) for i, x in enumerate(teacher_arr)]))
        self.student_mapping = self._join_dicts(student_mapping, dict([(x, i) for i, x in enumerate(student_arr)]))

    def _join_dicts(self, a: Dict, b: Dict) -> Dict:
        res = {}
        for key, value in a.items():
            res[key] = b[value[0]]
        return res

    def score(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> float:
        try:
            return self.scores[self.teacher_mapping[from_module]][self.student_mapping[to_module]]
        except KeyError:
            return 0



import torch

from iatransfer.toolkit.base_transfer import Transfer


class ClipTransfer(Transfer):
    """Clip tensor transfer algorithm for IAT.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def transfer_layer(self, tensor_from: torch.Tensor, tensor_to: torch.Tensor, *args, **kwargs) -> None:
        if tensor_from is None or tensor_to is None:
            return
        from_slices, to_slices = [], []
        for a, b in zip(tensor_from.shape, tensor_to.shape):
            if a < b:
                from_slices.append(slice(0, a))
                to_slices.append(slice((b - a) // 2, -((b - a + 1) // 2)))
            elif a > b:
                from_slices.append(slice((a - b) // 2, -((a - b + 1) // 2)))
                to_slices.append(slice(0, b))
            else:
                from_slices.append(slice(0, a))
                to_slices.append(slice(0, b))
        tensor_to[tuple(to_slices)] = tensor_from[tuple(from_slices)]




import re
from typing import List, Tuple, Any, Optional, Union

import torch
import torch.nn as nn

from iatransfer.toolkit.base_transfer import Transfer


class TraceTransfer(Transfer):
    """Trace tensor transfer algorithm for IAT.
    """

    def __init__(self, reverse_priority: bool = False, **kwargs) -> None:
        if reverse_priority:
            self.sgn = +1
        else:
            self.sgn = -1
        super().__init__(**kwargs)

    layers_classes = [
        nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
        nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
        nn.Linear
    ]
    layers_mapping = {}

    def create_layers_mapping(self, matched: List[Tuple[nn.Module, nn.Module]]):
        for matching in matched:
            if isinstance(matching, list):
                self.create_layers_mapping(matching)
            elif matching[0] is not None and matching[1] is not None:
                self.layers_mapping[matching[1]] = matching[0]

    class TorchSubstitute:
        """
            substitutes every call to torch library
        """

        def __getattr__(self, attr: Any) -> Any:
            def f(*args, **kwargs):
                return args[0] if len(args) > 0 else None

            return f

    class Module:

        def __init__(self, trace: 'torch.jit._trace.TracedModule', module: nn.Module,
                     outer_class: 'TraceTransfer') -> None:
            self.trace = trace
            self.module = module
            self.outer_class = outer_class

        def __getattr__(self, attr: Any) -> Any:
            if attr == "torch":
                return self.outer_class.TorchSubstitute()
            return self.__class__(getattr(self.trace, attr), getattr(self.module, attr), self.outer_class)

        def forward(self, ids: Optional[List[int]] = None):
            if len(list(self.module.children())) == 0 or any(
                    [isinstance(self.module, clazz) for clazz in self.outer_class.layers_classes]):
                # print("ids: ", ids)
                return self._transfer(self.module, ids)
            else:
                instructions = self.trace.code
                instructions = re.sub(r"forward[^\(]*\(", "forward(", instructions)
                instructions = instructions.replace("torch", "self.torch")
                instructions = instructions.split("\n")
                input_name = instructions[1].strip()[:instructions[1].strip().find(":")]
                exec(f"{input_name} = ids")
                i = 2
                while i < len(instructions) and instructions[i].find("return") == -1:
                    # print(instructions[i].strip(), "\n")
                    exec(instructions[i].strip())
                    i += 1
                if i < len(instructions):
                    return eval(re.search(r"return(.+)", instructions[i]).group(1).strip())
                return None

        def _get_center_slices(self, a: int, b: int) -> Tuple[slice, slice]:
            if a < b:
                return slice(0, a), slice((b - a) // 2, -((b - a + 1) // 2))
            elif a > b:
                return slice((a - b) // 2, -((a - b + 1) // 2)), slice(0, b)
            else:
                return slice(0, a), slice(0, b)

        def _get_output_channels(self, from_tensor: torch.Tensor, to_tensor: torch.Tensor) -> Union[
            Tuple[torch.Tensor, slice], Tuple[slice, slice]]:
            dims = [i for i in range(1, len(from_tensor.shape))]
            if from_tensor.shape[0] > to_tensor.shape[0]:
                if len(dims) == 0:
                    out_channels = list(enumerate(from_tensor))
                else:
                    out_channels = list(enumerate(from_tensor.abs().sum(dim=dims)))
                out_channels.sort(key=lambda w: self.outer_class.sgn * w[1])
                c = to_tensor.shape[0]
                from_ids = sorted([w[0] for w in out_channels[:c]])
                return torch.tensor(from_ids), slice(0, c)
            else:
                return self._get_center_slices(from_tensor.shape[0], to_tensor.shape[0])

        def _get_slices(self, from_tensor: torch.Tensor, to_tensor: torch.Tensor, ids: Optional[List[int]]) -> Tuple[
            Tuple[None, ...], Tuple[None, ...]]:
            n = len(from_tensor.shape)
            from_slices, to_slices = [None] * n, [None] * n
            if n > 1:  # use input ids to choose input channels
                a, b = from_tensor.shape[1], to_tensor.shape[1]
                # if ids is not None and a > b and len(ids) != b:
                #     print("HERE")
                if ids is not None and a > b and len(ids) == b and torch.max(ids) < from_tensor.shape[1]:
                    # print("ENTERED")
                    from_slices[1] = ids
                    to_slices[1] = slice(0, b)
                    from_tensor = from_tensor.index_select(1, ids)
                else:
                    from_slices[1], to_slices[1] = self._get_center_slices(a, b)
            # choose output channels
            if n > 1 or ids is None or len(ids) != to_tensor.shape[0] or to_tensor.shape[0] > from_tensor.shape[0]:
                from_slices[0], to_slices[0] = self._get_output_channels(from_tensor, to_tensor)
            else:
                from_slices[0], to_slices[0] = ids, slice(0, to_tensor.shape[0])
            # choose centers on the remaining channels
            for i in range(2, n):
                from_slices[i], to_slices[i] = self._get_center_slices(from_tensor.shape[i], to_tensor.shape[i])

            total_unsqueeze = 0
            for i in range(len(from_slices) - 1, -1, -1):
                if isinstance(from_slices[i], torch.Tensor):
                    # print(i)
                    for _ in range(total_unsqueeze):
                        from_slices[i] = from_slices[i].unsqueeze(-1)
                    total_unsqueeze += 1
            # print('total_unsqueeze: ', total_unsqueeze)

            return tuple(from_slices), tuple(to_slices)

        def _transfer(self, to_module: nn.Module, ids: Optional[List[int]]) -> Optional[List[int]]:
            if to_module not in self.outer_class.layers_mapping:
                return ids  # for example non-linearity - then pass ids
            from_module = self.outer_class.layers_mapping[to_module]
            if from_module is None:
                return ids
            # print("in: ", len(ids) if ids is not None else 0, ids)
            # print(to_module, from_module)

            from_slices, to_slices = self._get_slices(from_module.weight, to_module.weight, ids)
            # print(to_slices, from_slices)
            if to_module.weight is not None and from_module.weight is not None:
                to_module.weight[to_slices] = from_module.weight[from_slices]
            if to_module.bias is not None and from_module.bias is not None:
                to_ids = to_slices[0]
                from_ids = from_slices[0]
                if isinstance(to_ids, torch.Tensor):
                    to_ids = to_ids.flatten()
                if isinstance(from_ids, torch.Tensor):
                    from_ids = from_ids.flatten()
                to_module.bias[to_ids] = from_module.bias[from_ids]
            output_ids = from_slices[0]
            if isinstance(output_ids, slice):
                output_ids = torch.tensor([i for i in range(from_module.weight.shape[0])][output_ids])
            #     print("before out: ", [i for i in range(from_module.weight.shape[0])])
            #     print("slices: ", from_slices[0])
            # print("out: ", output_ids)
            return output_ids.flatten()

    def transfer(self, matched: List[Tuple[nn.Module, nn.Module]], *args, **kwargs) -> None:
        to_module = args[0] if len(args) > 0 else kwargs['context']['to_module']
        with torch.no_grad():
            self.create_layers_mapping(matched)
            to_module.eval()
            module = self.Module(torch.jit.trace(to_module, torch.randn(1, 3, 300, 300)), to_module, self)
            module.forward()

    def transfer_layer(self, tensor_from: torch.Tensor, tensor_to: torch.Tensor, *args, **kwargs) -> None:
        raise ValueError("Not implemented")



import json


def read_contents(filename: str, encoding: str = 'utf-8') -> str:
    with open(filename, 'r', encoding=encoding) as file:
        return file.read()


def read_json(filename: str) -> dict:
    with open(filename) as json_file:
        return json.load(json_file)



from abc import ABC, abstractmethod
from typing import List, Tuple

import torch
import torch.nn as nn


class Transfer(ABC):
    """Base-class for any tensor transfer algorithm detectable from IAT.
    """

    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs

    def transfer(self, matched_tensors: List[Tuple[nn.Module, nn.Module]], *args, **kwargs) -> None:
        """Invokes the transfer procedure for a list of matched tensors.
        """
        with torch.no_grad():
            for matching in matched_tensors:
                if isinstance(matching, list):
                    self.transfer(matching)
                elif matching[0] is not None and matching[1] is not None:
                    self.transfer_layer(matching[0].weight, matching[1].weight)
                    self.transfer_layer(matching[0].bias, matching[1].bias)

    @abstractmethod
    def transfer_layer(self, tensor_from: nn.Module, tensor_to: nn.Module, *args, **kwargs) -> None:
        """Transfers the matched layers. Changes the contents of matched tensors in model_to.
        """
        pass

    def __call__(self, matched_tensors: List[Tuple[nn.Module, nn.Module]], *args, **kwargs) -> None:
        """Alias for 'transfer'.
        """
        self.transfer(matched_tensors, *args, **kwargs)



import collections
import os
import random
import sys
from copy import copy
from typing import Dict, List

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import torch
import torch.nn as nn
from graphviz import Digraph
from karateclub import FeatherNode, NetMF
from networkx.drawing.nx_agraph import graphviz_layout
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPRegressor
from torch.autograd import Variable

use_cuda = torch.cuda.is_available()
device = torch.device("cuda" if use_cuda else "cpu")


def apply_tli(model, teacher=None):
    model_teacher = str_to_model(teacher)
    transfer(model_teacher, model)
    return model


def get_tli_score(model_from, model_to):
    model_a = str_to_model(model_from)
    model_b = str_to_model(model_to)
    sim, _, _, _ = transfer(model_a, model_b)
    return sim


def get_model_timm(name="dla46x_c"):
    try:
        import timm
    except:
        raise Exception("timm package is not installed! try `pip install timm`")

    model = timm.create_model(name, num_classes=10, in_chans=3, pretrained=True)
    return model


def str_to_model(name):
    if isinstance(name, str):
        print(f"loading `{name}` from pytorch-image-models...")
        model = get_model_timm(name)
    else:
        model = name
    return model


def apply_hard_reset(model):
    for layer in model.modules():
        if hasattr(layer, "reset_parameters"):
            nn.init.zeros_(layer.weight)
            if layer.bias is not None:
                nn.init.zeros_(layer.bias)
    return model


def fn_inject(from_tensor, to_tensor):
    from_slices, to_slices = [], []
    for a, b in zip(from_tensor.shape, to_tensor.shape):
        if a < b:
            from_slices.append(slice(0, a))
            to_slices.append(slice((b - a) // 2, -((b - a + 1) // 2)))
        elif a > b:
            from_slices.append(slice((a - b) // 2, -((a - b + 1) // 2)))
            to_slices.append(slice(0, b))
        else:
            from_slices.append(slice(0, a))
            to_slices.append(slice(0, b))
    to_tensor[tuple(to_slices)] = from_tensor[tuple(from_slices)]


def get_networkx(edges, dag=True):
    if dag:
        G = nx.DiGraph()
    else:
        G = nx.Graph()
    G.add_edges_from(edges)
    return G


def show_networkx(graph):
    if isinstance(graph, list):
        graph = get_networkx(edges=graph)
    pos = graphviz_layout(graph, prog="dot")
    nx.draw(graph, pos, with_labels=True, arrows=True)
    plt.show()


def dag_split(edges, token, root=None):
    graph = {}
    for a, b in edges:
        if a not in graph:
            graph[a] = []
        if b not in graph:
            graph[b] = []
        graph[a].append(b)
        graph[b].append(a)
    edges_split = []
    visited, queue = set(), collections.deque([root])
    while queue:
        stop = False
        node_root = queue.popleft()
        if node_root not in graph:
            continue
        if node_root == token:
            break
        for node in graph[node_root]:
            if node not in visited:
                if node == token:
                    stop = True
                edges_split.append([node_root, node])
                visited.add(node)
                queue.append(node)
        if stop:
            break

    if not edges_split:
        edges_split.append([token, token])
    return edges_split


def graph_splits(edges, nodes=False):
    G = get_networkx(edges)
    order = list(nx.topological_sort(G))
    if len(order) == 0:
        return {}
    idx_src, idx_dst = order[0], order[-1]
    if not nodes:
        nodes = set()
        for a, b in edges:
            nodes.add(a)
            nodes.add(b)
    split_map = {}
    for idx in nodes:
        in_tree = dag_split(edges, idx, root=idx_src)
        out_tree = dag_split(edges, idx, root=idx_dst)
        split_map[idx] = {"in-tree": in_tree, "out-tree": out_tree}
    return split_map


def graph_norm(edges, attr=None):
    normal_id_map = {}
    normal_id_iter = [0]
    rev_mask = {}

    def __for_single(idx):
        if not idx in normal_id_map:
            normal_id_map[idx] = normal_id_iter[0]
            rev_mask[normal_id_iter[0]] = idx
            normal_id_iter[0] += 1

    for a, b in edges:
        __for_single(a)
        __for_single(b)

    norm_edges = []
    for a, b in edges:
        norm_edges.append([normal_id_map[a], normal_id_map[b]])

    norm_attr = []
    if attr:
        for i in range(len(normal_id_map.keys())):
            norm_attr.append(attr[rev_mask[i]])

    return norm_edges, rev_mask, norm_attr


def utils_map_to_mask(split_map):
    mask, graphs = [], []
    for key, split_dict in split_map.items():
        for dict_key in split_dict.keys():
            _g, rev_mask, _ = graph_norm(split_dict[dict_key])
            g = get_networkx(_g, dag=False)
            mask.append([key, dict_key])
            graphs.append(g)
    return mask, graphs


def utils_mask_to_map(mask, X):
    split_map = {}
    for i, (key, dict_key) in enumerate(mask):
        if key not in split_map:
            split_map[key] = {}
        split_map[key][dict_key] = X[i]
    return split_map


def split_flow_level(graph):
    edges = []
    for edge in graph.cluster_links:
        cluster_idx_1 = graph.nodes[edge[0]].cluster_idx
        cluster_idx_2 = graph.nodes[edge[1]].cluster_idx
        edges.append([cluster_idx_1, cluster_idx_2])
    return graph_splits(edges)


def split_cluster_level(graph, cluster_idx):
    edges = graph.cluster_map[cluster_idx].edges
    return graph_splits(edges)


def encode_graph(split_map):
    mask, graphs = utils_map_to_mask(split_map)

    from karateclub import GL2Vec

    model = GL2Vec(dimensions=16)
    print("FIT")
    model.fit(graphs)
    print("EMBEDDING")
    X = model.get_embedding()
    print("-------------------->", X.shape)

    return utils_mask_to_map(mask, X)


class TLIConfig(object):
    def __init__(self, adict):
        self.__dict__.update(adict)


embedding_dim = 5
CONFIG = TLIConfig(
    {

        "node_embedding_attributed": FeatherNode(
            eval_points=4, order=4, svd_iterations=100, reduction_dimensions=32
        ),
        "node_embedding_neighbourhood": NetMF(
            dimensions=embedding_dim
        ),

        "autoencoder": MLPRegressor(
            max_iter=100,
            early_stopping=False,
            activation="relu",
            solver="adam",
            tol=0.0001,

            hidden_layer_sizes=(200, 50, 25,),
            warm_start=True,
            learning_rate_init=0.0005,
            alpha=0.001,
            verbose=True,
        ),
        "test_size": 0.05,
        "samples_per_tensor": 10,
    }
)


def E_nodes(edges, attr=None):
    norm_graph, rev_mask, norm_attr = graph_norm(edges, attr=attr)

    if len(rev_mask) == 0:
        return []

    model = (
        CONFIG.node_embedding_attributed
        if attr
        else CONFIG.node_embedding_neighbourhood
    )

    graph = get_networkx(norm_graph, dag=False)
    if attr:
        model.fit(graph, np.array(norm_attr))
        X = model.get_embedding()
    else:
        model.fit(graph)
        X = model.get_embedding()

    print(f"[E_nodes {X.shape}]", end="")

    encoded_nodes = {}
    for i in range(X.shape[0]):
        encoded_nodes[rev_mask[i]] = X[i]
    return encoded_nodes


def F_architecture(graph, mlb=None, mfa=None):
    edges = []
    cluster_feature = {}
    for cluster_idx, cluster in graph.cluster_map.items():
        cluster_feature[cluster_idx] = [len(cluster.nodes) / (1 + len(cluster.edges))]
    for edge in graph.cluster_links:
        cluster_idx_1 = graph.nodes[edge[0]].cluster_idx
        cluster_idx_2 = graph.nodes[edge[1]].cluster_idx
        edges.append([cluster_idx_1, cluster_idx_2])
    P = E_nodes(edges, attr=cluster_feature)

    S = {}
    for cluster_idx in graph.cluster_map.keys():
        edges = graph.cluster_map[cluster_idx].edges

        if len(edges) > embedding_dim:
            obj = E_nodes(edges)
        else:
            obj = {}
            for idx in graph.cluster_map[cluster_idx].nodes:
                obj[idx] = np.array([0.0] * embedding_dim)
        S.update(obj)

    N = {}
    vec = []
    for idx, node in graph.nodes.items():
        vec.append(__encode(node.name))

    vec = mlb.transform(vec)
    vec = mfa.transform(vec)
    vec_final = []
    for i, (idx, node) in enumerate(
            graph.nodes.items()
    ):
        _shape4 = nn.ConstantPad1d((0, 4 - len(node.size)), 0.0)(
            torch.tensor(node.size)
        )

        shape4 = _shape4.type(torch.FloatTensor) / torch.max(1 + _shape4)
        if shape4[0] > shape4[1]:
            rot = 1
        else:
            rot = 0
        _idx_rev = (graph.max_idx - node.idx) / graph.max_idx
        _idx_rev2 = (node.idx) / graph.max_idx
        _level_rev = (graph.max_level - node.level) / graph.max_level
        _level_rev2 = (node.level) / graph.max_level
        _cluster_rev = (graph.max_idx - node.cluster_idx) / graph.max_idx
        _cluster_rev2 = (node.cluster_idx) / graph.max_idx
        _type = 0 if ".bias" in node.name else 1

        vec_final.append(np.array(
            [rot]
            + shape4.tolist()
            + [(_idx_rev + _cluster_rev + _level_rev) / 3,
               (_idx_rev2 + _cluster_rev2 + _level_rev2) / 3, _type]
        ))

    from sklearn import preprocessing

    _pp = preprocessing.StandardScaler()

    vec_final = _pp.fit_transform(vec_final)

    for i, (idx, node) in enumerate(
            graph.nodes.items()
    ):
        N[idx] = np.array(vec_final[i].tolist() + vec[i].tolist())

    print("(encode_graph ended)")
    return P, S, N


def __q(a, b):
    return np.array(a) + np.array(b)


def __shape_score(s1, s2):
    if len(s1) != len(s2):
        return 0
    score = 1
    for x, y in zip(s1, s2):
        score *= min(x / y, y / x)
    return score


def gen_dataset(graph, P, S, N, EG, prefix=""):
    X, y = [], []

    for idx, node in graph.nodes.items():
        if node.type != "W":
            continue

        cluster_idx = node.cluster_idx

        for _ in range(CONFIG.samples_per_tensor):
            p_src = np.array(P[cluster_idx])
            r = np.random.uniform(low=-0.05, high=0.05, size=p_src.shape)
            p_src += r
            s_src = np.array(S[idx])
            r = np.random.uniform(low=-0.05, high=0.05, size=s_src.shape)
            s_src += r
            q_src = p_src.tolist() + s_src.tolist() + list(N[idx]) + \
                    EG[f"{prefix}_{idx}"]["in-tree"].tolist()
            X.append(__q(q_src, q_src))

            y.append(1 + np.random.uniform(low=-0.05, high=0.05))

        q_src = list(P[cluster_idx]) + list(S[idx]) + list(N[idx]) + \
                EG[f"{prefix}_{idx}"]["in-tree"].tolist()

        X.append(__q(q_src, q_src))
        y.append(1)

        def __get_node(cluster_idx=None, type=None):
            r_idx = None
            if cluster_idx is not None:
                nodes = list(graph.cluster_map[cluster_idx].nodes)
            else:
                nodes = list(graph.nodes.keys())
            for _ in range(len(N)):
                r_idx = random.choice(nodes)
                if graph.nodes[r_idx].type == type or not type:
                    break
            return r_idx

        for _ in range(CONFIG.samples_per_tensor):
            r_idx = __get_node(cluster_idx=cluster_idx, type="W")
            r_cluster_idx = cluster_idx
            if idx == r_idx:
                continue

            q_dst = list(P[r_cluster_idx]) + list(S[r_idx]) + list(N[r_idx]) + \
                    EG[f"{prefix}_{r_idx}"]["in-tree"].tolist()

            N_bonus = 0
            N_dist = np.linalg.norm(N[idx] - N[r_idx])

            if N_dist <= 1:
                N_bonus = (1 - N_dist) / 4

            X.append(__q(q_src, q_dst))
            y.append(
                N_bonus
                + 0.25
                + 0.5 * __shape_score(graph.nodes[idx].size, graph.nodes[r_idx].size)
            )

        for _ in range(CONFIG.samples_per_tensor):
            r_idx = __get_node(cluster_idx=None, type="W")
            r_cluster_idx = graph.nodes[r_idx].cluster_idx
            if r_cluster_idx == cluster_idx:
                continue
            if idx == r_idx:
                continue

            q_dst = list(P[r_cluster_idx]) + list(S[r_idx]) + list(N[r_idx]) + \
                    EG[f"{prefix}_{r_idx}"]["in-tree"].tolist()

            N_bonus = 0
            N_dist = np.linalg.norm(N[idx] - N[r_idx])

            if N_dist <= 1:
                N_bonus = (1 - N_dist) / 4

            S_bonus = 0
            S_dist = np.linalg.norm(S[idx] - S[r_idx])

            if S_dist <= 1:
                S_bonus = (1 - S_dist) / 4

            X.append(__q(q_src, q_dst))
            y.append(
                N_bonus / 2
                + S_bonus / 2
                + 0.25 * __shape_score(graph.nodes[idx].size, graph.nodes[r_idx].size)
            )

    print("DATASET", np.array(X).shape)

    return X, y


def __encode(x):
    x = x.replace(".weight", "").replace(".bias", "")
    x = x.replace("blocks", "")
    if "Backward" in x:
        x = ""

    _vec = list(x)

    _lvl = [s for s in _vec if s.isdigit()]
    _lvl = "".join(_lvl)
    _vec = list(set(_vec))
    if _lvl:
        _vec.append(_lvl)

    return _vec


def score_autoencoder(graph_src, graph_dst):
    from sklearn.preprocessing import MultiLabelBinarizer
    from sklearn.manifold import Isomap

    mlb = MultiLabelBinarizer()

    vec = []

    for idx, node in graph_dst.nodes.items():
        vec.append(__encode(node.name))

    mlb.fit(vec)
    _l1 = len(graph_dst.nodes.keys())
    _l2 = len(graph_dst.cluster_map.keys())

    mfa = Isomap(n_components=min(_l1 // 2, 30), n_neighbors=min(_l1 // 10, 50), p=3)
    _vec = mlb.transform(vec)
    mfa.fit(_vec)

    P_src, S_src, N_src = F_architecture(graph_src, mlb=mlb, mfa=mfa)
    P_dst, S_dst, N_dst = F_architecture(graph_dst, mlb=mlb, mfa=mfa)

    split_map = {}
    for cluster_idx in graph_src.cluster_map.keys():
        _split_map = split_cluster_level(graph_src, cluster_idx)
        for key in _split_map:
            split_map[f"src_{key}"] = _split_map[key]
    print("(graph_src ended)")
    for cluster_idx in graph_dst.cluster_map.keys():
        _split_map = split_cluster_level(graph_dst, cluster_idx)
        for key in _split_map:
            split_map[f"dst_{key}"] = _split_map[key]
    print("(graph_dst ended)")
    EG = encode_graph(split_map)

    X1, y1 = gen_dataset(graph_src, P_src, S_src, N_src, EG, prefix="src")
    X2, y2 = gen_dataset(graph_dst, P_dst, S_dst, N_dst, EG, prefix="dst")
    X = X1 + X2
    y = y1 + y2

    print("DATASET FULL", np.array(X).shape)

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=CONFIG.test_size, random_state=42
    )

    model = copy(CONFIG.autoencoder)

    model.fit(X_train, y_train)

    y_hat = model.predict(X_test)
    loss = mean_squared_error(y_test, y_hat)
    print(f" LOSS --> {loss}")

    def __norm_weights(graph):
        arr, imap, i = [], {}, 0
        for _, (idx, node) in enumerate(graph.nodes.items()):
            if node.type != "W":
                continue
            arr.append(idx)
            imap[idx] = i
            i += 1
        return arr, imap

    src_arr, src_map = __norm_weights(graph_src)
    dst_arr, dst_map = __norm_weights(graph_dst)

    n, m = len(src_arr), len(dst_arr)
    scores = np.zeros((n, m))

    for dst_j, idx_dst in enumerate(dst_arr):
        node_dst = graph_dst.nodes[idx_dst]
        dst_type = node_dst.name.split(".")[-1]

        q_dst = (
                list(P_dst[node_dst.cluster_idx])
                + list(S_dst[idx_dst])
                + list(N_dst[idx_dst])
                + list(EG[f"dst_{idx_dst}"]["in-tree"].tolist())
        )

        q_arr = []
        for src_i, idx_src in enumerate(src_arr):
            node_src = graph_src.nodes[idx_src]
            src_type = node_src.name.split(".")[-1]

            q_src = (
                    list(P_src[node_src.cluster_idx])
                    + list(S_src[idx_src])
                    + list(N_src[idx_src])
                    + list(EG[f"src_{idx_src}"]["in-tree"].tolist())
            )
            q_arr.append(__q(q_src, q_dst))
            scores[src_i, dst_j] = __shape_score(node_dst.size, node_src.size)

            if dst_type != src_type:
                scores[src_i, dst_j] = 0

        y_hat = model.predict(q_arr)
        scores[:, dst_j] *= y_hat

    return scores, src_arr, dst_arr


def transfer(model_src, model_dst=None, teacher=None, inject=True, debug=False):
    if model_src and model_dst:

        print("API: V2")
        pass
    elif not model_dst and teacher:

        print("API: V1")
        model_src, model_dst = teacher, model_src
    else:
        raise Exception("where is teacher?! is this a joke?")

    graph_src = get_graph(model_src)
    graph_dst = get_graph(model_dst)

    if debug:
        show_graph(graph_src, ver=3, path="__tli_src")
        show_graph(graph_dst, ver=3, path="__tli_dst")

    scores, src_arr, dst_arr = score_autoencoder(graph_src, graph_dst)

    remap = {}
    n, m = len(src_arr), len(dst_arr)

    beta = 0.5
    smap = copy(scores)
    for _ in range(n * m):
        i, j = np.unravel_index(smap.argmax(), smap.shape)
        smap[i, :] *= beta

        if dst_arr[j] not in remap:
            smap[:, j] = 0
            remap[dst_arr[j]] = src_arr[i]

    window_size = 0.25
    for _dst_j, idx_dst in enumerate(dst_arr[::-1]):
        dst_j = m - _dst_j - 1
        ith = dst_j / m
        shift = max(int(ith * n - window_size * n), 0)
        i = np.argmax(smap[shift:, dst_j]) + shift
        if idx_dst not in remap:
            remap[idx_dst] = src_arr[i]

    window_size = 1
    for _dst_j, idx_dst in enumerate(dst_arr[::-1]):
        dst_j = m - _dst_j - 1
        ith = dst_j / m
        shift = max(int(ith * n - window_size * n), 0)
        i = np.argmax(smap[shift:, dst_j]) + shift
        if idx_dst not in remap:
            remap[idx_dst] = src_arr[i]

    seen = set()
    all_scores = []
    error_n, error_sum = 0, 0
    print(" " * 45 + "[[src]]" + " " * 30 + "[[dst]]")
    for j, idx_dst in enumerate(dst_arr):
        node_dst = graph_dst.nodes[idx_dst]

        idx_src = remap[idx_dst]
        score = scores[src_arr.index(idx_src), j]
        all_scores.append(score)

        name_src = graph_src.nodes[idx_src].name
        name_dst = node_dst.name
        color_code = "\x1b[1;37;40m"
        if name_src != name_dst:
            error_sum += 1
            color_code = "\x1b[1;31;40m"

        color_end = "\x1b[0m"
        print(
            f"src= {idx_src:3} | dst= {idx_dst:3} | "
            + f"S= {round(score, 2):4} | {color_code}{name_src:30}{color_end} / "
            + f"{name_dst:10}"
        )

        seen.add(idx_src)
        error_n += 1

    sim = max(0, min(1, np.mean(all_scores)))

    print("=== MATCH =================")
    n = len(graph_src.nodes.keys())
    print(f"  SIM --> \x1b[0;34;40m{round(sim, 4)}\x1b[0m")
    print(f" SEEN --> {len(seen):5} / {n:5} | {round(len(seen) / n, 2)}")
    print(f"ERROR --> {error_sum:5} / {error_n:5} | {round(error_sum / error_n, 2)}")
    print("===========================")

    if debug:
        show_remap(graph_src, graph_dst, remap, path="__tli_remap")

    if inject:
        p_src_ref = {}
        for name, param in model_src.named_parameters():
            p_src_ref[name] = param
        p_dst_ref = {}
        for name, param in model_dst.named_parameters():
            p_dst_ref[name] = param

        with torch.no_grad():
            for idx_dst, idx_src in remap.items():
                node_src = graph_src.nodes[idx_src]
                node_dst = graph_dst.nodes[idx_dst]
                p_src = p_src_ref[node_src.name]
                p_dst = p_dst_ref[node_dst.name]
                fn_inject(p_src, p_dst)

    return sim, remap, graph_src, graph_dst


class Node:
    def __init__(self):
        self.idx = 0
        self.var = None
        self.type = None
        self.size = ()
        self.level = 1
        self.cluster_idx = 1


class Graph:
    def __init__(self):
        self.nodes = None
        self.edges = None

        self.cluster_map = None
        self.cluster_links = None

        self.max_level = None
        self.max_idx = None


class Cluster:
    def __init__(self):
        self.cluster_idx = 0
        self.nodes = []
        self.edges = []


def make_graph(var, params=None) -> Graph:
    graph = Graph()
    mod_op = ["AddBackward0", "MulBackward0", "CatBackward"]

    if params is not None:
        assert all(isinstance(p, Variable) for p in params.values())
        param_map = {id(v): k for k, v in params.items()}

    def __get_type(var):
        node = Node()
        node.var = var
        if hasattr(var, "variable"):
            u = var.variable
            node_name = param_map[id(u)]
            size = list(u.size())
            node.name = node_name
            node.size = size
            node.type = "W"
        else:
            node_name = str(type(var).__name__)
            if node_name in mod_op:
                node.type = "OP"
            else:
                node.type = "F"
            node.name = node_name
        return node

    normal_id_map = {}
    normal_id_iter = [0]

    def __normal_id(var):
        __pointer_idx = id(var)
        if __pointer_idx in normal_id_map:
            return normal_id_map[__pointer_idx]
        else:
            normal_id_map[__pointer_idx] = normal_id_iter[0]
            normal_id_iter[0] += 1
            return normal_id_iter[0] - 1

    def __bfs(graph, degree=2):
        nodes = {}
        edges = {}
        _rev_edges = {}
        _level_map = {}
        _mod_op_map = {}
        visited, queue = set(), collections.deque([graph])
        while queue:
            var = queue.popleft()
            idx_root = __normal_id(var)
            if idx_root not in _level_map:
                _level_map[idx_root] = 1
            if idx_root not in _mod_op_map:
                _mod_op_map[idx_root] = idx_root
            if idx_root not in nodes:
                nodes[idx_root] = __get_type(var)
                nodes[idx_root].cluster_idx = idx_root
                nodes[idx_root].type = "OP"
            if idx_root not in edges:
                edges[idx_root] = []
            if idx_root not in _rev_edges:
                _rev_edges[idx_root] = []
            for _u in var.next_functions:
                u = _u[0]
                idx = __normal_id(u)
                if not u:
                    continue
                edges[idx_root].append(idx)
                if idx not in _rev_edges:
                    _rev_edges[idx] = []
                _rev_edges[idx].append(idx_root)
                if u not in visited:
                    _level_map[idx] = _level_map[idx_root] + 1
                    node = __get_type(u)
                    node.idx = idx
                    if node.type == "OP":
                        _mod_op_map[idx] = idx_root
                    else:
                        _mod_op_map[idx] = _mod_op_map[idx_root]
                    node.level = _level_map[idx]
                    node.cluster_idx = _mod_op_map[idx]
                    nodes[idx] = node

                    visited.add(u)
                    queue.append(u)

        if degree:
            visited, queue = set(), collections.deque([graph])
            for idx_root in _rev_edges:

                if len(_rev_edges[idx_root]) >= degree \
                        and nodes[idx_root].type != "W":
                    nodes[idx_root].type = "OP"
            while queue:
                var = queue.popleft()
                idx_root = __normal_id(var)
                for _u in var.next_functions:
                    u = _u[0]
                    idx = __normal_id(u)
                    if not u:
                        continue
                    if u not in visited:
                        node = nodes[idx]
                        if node.type == "OP":
                            _mod_op_map[idx] = idx_root
                        else:
                            _mod_op_map[idx] = _mod_op_map[idx_root]
                        node.cluster_idx = _mod_op_map[idx]
                        nodes[idx] = node
                        visited.add(u)
                        queue.append(u)
        max_level = 0
        for _, node_level in _level_map.items():
            max_level = max(max_level, node_level)
        return nodes, edges, max_level

    if isinstance(var, tuple):
        raise Exception("Lord Dark Tensor: have not implemented that feature")
        sys.exit(1)
        for v in var:
            __bfs(v.grad_fn)
    else:

        nodes, edges, max_level = __bfs(var.grad_fn)

    graph.nodes = nodes
    graph.edges = edges

    graph.cluster_map, graph.cluster_links = make_clusters(graph)
    if len(graph.cluster_map.keys()) <= 1:
        graph.cluster_links.append([0, 0])

    graph.max_level = max_level
    graph.max_idx = normal_id_iter[0]

    return graph


def make_clusters(graph):
    cluster_map = {}
    cluster_links = []
    for idx, node in graph.nodes.items():
        if node.cluster_idx not in cluster_map:
            cluster_map[node.cluster_idx] = Cluster()
        cluster_map[node.cluster_idx].nodes.append(idx)
    for idx_root, edges in graph.edges.items():
        node_root = graph.nodes[idx_root]
        for idx in edges:
            if graph.nodes[idx].type == "OP":
                cluster_links.append([idx, idx_root])
                continue
            cluster_map[node_root.cluster_idx].edges.append([idx, idx_root])
    return cluster_map, cluster_links


def get_graph(model, input=None):
    graph = None
    input_shape = [input] if input else [(3, 32, 32), (1, 31, 31), (3, 224, 224)]
    for _input_shape in input_shape:
        x = torch.randn(32, *_input_shape)
        try:
            x = x.to(device)
            model = model.to(device)
            graph = make_graph(model(x), params=dict(model.named_parameters()))
            break
        except Exception as err:
            print("ERROR", err)
            continue
    if not graph:
        raise Exception("something really wrong!")
    return graph


def get_idx_to_layers_mapping(model: nn.Module, graph: Graph) -> Dict[int, nn.Module]:
    names_to_layers_mapping = {}

    def dfs(model: nn.Module, name_prefix: List[str]):
        for child_name, child in model.named_children():
            dfs(child, name_prefix + [child_name])
        names_to_layers_mapping[".".join(name_prefix)] = model

    dfs(model, [])

    ids_to_layers_mapping = {}
    for node in graph.nodes.values():
        if node.type == "W":
            node_name = node.name.replace(".weight", "").replace(".bias", "")
            layer = names_to_layers_mapping[node_name]
            ids_to_layers_mapping[node.idx] = layer

    return ids_to_layers_mapping


def make_dot(graph, ver=0, prefix="", rankdir="TB"):
    graph_idx = id(graph)

    node_attr = dict(
        style="filled",
        shape="box",
        align="left",
        fontsize="12",
        ranksep="0.1",
        height="0.2",

    )

    graph_attr = dict(
        rank="same",

        rankdir=rankdir,

    )

    print(f"graph_idx={graph_idx}")
    graph_name = f"cluster_{graph_idx}"
    dot = Digraph(name=graph_name, node_attr=node_attr, graph_attr=graph_attr)

    cluster_map, cluster_links = graph.cluster_map, graph.cluster_links

    def __show_graph_nodes():
        for idx, node in graph.nodes.items():
            _header_name = (
                    f"[c = {node.cluster_idx} / "
                    + f"l = {node.level} / "
                    + f"idx = {node.idx}]\n{node.name}"
            )
            if node.type == "OP":
                dot.node(prefix + str(idx), _header_name, fillcolor="green")
            elif node.type == "W":
                dot.node(
                    prefix + str(idx),
                    _header_name + f"\n{node.size}",
                    fillcolor="lightblue",
                )
            else:
                dot.node(prefix + str(idx), _header_name)

    def __show_graph_edges():
        for idx_root, edges in graph.edges.items():
            for idx in edges:
                dot.edge(prefix + str(idx), prefix + str(idx_root), color="black")

    def __show_clusters():
        for cluster_idx, cluster in cluster_map.items():
            with dot.subgraph(name=f"cluster_{graph_idx}_{cluster_idx}") as c:
                c.attr(style="filled", color="lightgrey")
                for edge in cluster.edges:
                    c.edge(prefix + str(edge[0]), prefix + str(edge[1]), color="black")
                c.attr(label=f"cluster {cluster_idx}")
                if rankdir == "LR":
                    c.attr(rotate="90", rankdir="LR")

    if ver == 0:
        __show_graph_nodes()
        __show_graph_edges()

    if ver == 1:
        cluster_seen = set()

        for idx, node in graph.nodes.items():
            if node.type == "OP" and node.cluster_idx not in cluster_seen:
                nodes_in_cluster = len(cluster_map[node.cluster_idx].nodes)
                name = f"{node.cluster_idx} ({nodes_in_cluster})"
                dot.node(prefix + str(node.cluster_idx), name, fillcolor="orange")
                cluster_seen.add(node.cluster_idx)

        for edge in cluster_links:
            cluster_idx_1 = graph.nodes[edge[0]].cluster_idx
            cluster_idx_2 = graph.nodes[edge[1]].cluster_idx
            dot.edge(
                prefix + str(cluster_idx_1),
                prefix + str(cluster_idx_2),
                color="darkgreen",
                penwidth="3",
            )

    if ver == 2:
        __show_clusters()

    if ver == 3:
        __show_graph_nodes()

        for edge in cluster_links:
            dot.edge(
                prefix + str(edge[0]),
                prefix + str(edge[1]),
                color="darkgreen",
                minlen="3",
                penwidth="3",
            )

        __show_clusters()

    resize_dot(dot)
    dot.engine = "dot"
    return dot


def resize_dot(dot, size_per_element=0.15, min_size=12):
    num_rows = len(dot.body)
    content_size = num_rows * size_per_element
    size = max(min_size, content_size)
    size_str = str(size) + "," + str(size)
    dot.graph_attr.update(size=size_str)
    return size


def show_graph(model, ver=0, path="__tli_debug", input=None):
    if not isinstance(model, Graph):
        graph = get_graph(model, input=input)
    else:
        graph = model
    dot = make_dot(graph, ver=ver, prefix="this")
    dot.render(filename=path)
    os.system(f"rm {path}")
    print("saved to file")


def show_remap(g1, g2, remap, path="__tli_debug", for_edges=False):
    dot_g1 = make_dot(g1, ver=3, prefix="src", rankdir="TB")
    dot_g2 = make_dot(g2, ver=3, prefix="dst", rankdir="LR")

    graph_attr = dict(rankdir="TB", )
    dot = Digraph(name="root", graph_attr=graph_attr)
    dot_g2.graph_attr.update(rotate="90")

    dot_g2.graph_attr.update(compound="True")
    dot_g1.graph_attr.update(compound="True")
    dot.graph_attr.update(compound="True")
    dot.subgraph(dot_g2)
    dot.subgraph(dot_g1)
    from matplotlib.colors import to_hex
    import matplotlib.pyplot as plt

    cmap = plt.get_cmap("gist_rainbow")
    if not for_edges:
        arr = g1.cluster_map.keys()
    else:
        arr = range(len(remap.keys()))

    colors = cmap(np.linspace(0, 1, len(arr)))
    colors_map = {}
    for (i, color) in zip(arr, colors):
        colors_map[i] = color

    for i, (idx_dst, idx_src) in enumerate(remap.items()):
        if not for_edges:
            color = colors_map[g1.nodes[idx_src].cluster_idx]
        else:
            color = colors_map[i]
        dot.edge(
            "src" + str(idx_src),
            "dst" + str(idx_dst),
            color=to_hex(color),

            constraint="false",
            penwidth="5",
            weight="5",
        )
    dot.render(filename=path)
    os.system(f"rm {path}")
    print("saved to file")


if __name__ == "__main__":
    if True:
        from research_models import get_model_debug, ResNetUNet

        model_debug_small = get_model_debug(seed=42, channels=3, classes=10)
        model_debug_large = get_model_debug(seed=3, channels=3, classes=10)
        model_unet = ResNetUNet(n_class=6)

        show_graph(model_debug_small, ver=0, path="__tli_figure_1_all")
        show_graph(model_debug_small, ver=3, path="__tli_figure_1_graph")

        transfer(model_debug_small, model_debug_large, debug=True)

        show_graph(model_unet, ver=1, path="__tli_figure_unet")

    if False:
        model_A = get_model_timm("efficientnet_lite1")
        model_B = get_model_timm("mnasnet_100")

    if False:
        model_A = get_model_timm("efficientnet_lite1")
        model_B = get_model_timm("efficientnet_lite0")

    if True:
        model_A = get_model_timm("efficientnet_lite0")
        model_B = get_model_timm("efficientnet_lite1")

    if False:
        model_A = get_model_timm("efficientnet_lite1")
        model_B = get_model_timm("efficientnet_lite1")

    if False:
        model_A = get_model_timm("efficientnet_lite0")
        model_B = get_model_timm("efficientnet_lite0")

    if False:
        model_A = get_model_timm("mixnet_s")
        model_B = get_model_timm("mixnet_s")

    if False:
        model_A = get_model_timm("mixnet_s")
        model_B = get_model_timm("mixnet_m")

    if False:
        model_A = get_model_timm("mixnet_m")
        model_B = get_model_timm("mixnet_s")

    if False:
        model_A = get_model_timm("efficientnet_lite1")
        model_B = get_model_timm("tf_efficientnet_b0_ap")

    if False:
        model_A = get_model_timm("tf_efficientnet_b0_ap")
        model_B = get_model_timm("mnasnet_100")

    if False:
        model_A = get_model_timm("mixnet_s")
        model_B = get_model_timm("mnasnet_100")

    if False:
        model_A = get_model_timm("regnetx_002")
        model_B = get_model_timm("efficientnet_lite0")

    transfer(model_A, model_B, debug=True)



from typing import Tuple, List, Union

import networkx as nx
import numpy as np
import torch.nn as nn

from iatransfer.toolkit.matching.dp_matching import DPMatching


class BipartiteMatching(DPMatching):
    """Bipartite matching algorithm for IAT.
    """

    def match(self, from_module: List[Union[nn.Module, List[nn.Module]]],
              to_module: List[Union[nn.Module, List[nn.Module]]], *args, **kwargs) \
            -> List[Tuple[nn.Module, nn.Module]]:
        _, matched, _ = self._match_bipartite(from_module, to_module)
        return matched

    def sim(self, from_module: List[Union[nn.Module, List[nn.Module]]],
            to_module: List[Union[nn.Module, List[nn.Module]]]):
        score, _, _ = self._match_bipartite(from_module, to_module)
        return score

    def _match_bipartite(self, flat_from_module: List[Union[nn.Module, List[nn.Module]]],
                         flat_to_module: List[Union[nn.Module, List[nn.Module]]]) \
            -> Tuple[float, List[Tuple[nn.Module, nn.Module]], List[Tuple[int, int]]]:
        m = len(flat_from_module)
        n = len(flat_to_module)
        scores = np.zeros((n, m))

        graph = nx.Graph()

        for i in range(n):
            for j in range(m):
                dist_score = 1 - (abs(i / n - j / m) ** 2)
                scores[i, j] = dist_score * self._compute_score(flat_from_module[j], flat_to_module[i])
                graph.add_edge(f'x{i}', f'{j}', weight=scores[i, j])

        max_scores = np.max(scores, axis=0).sum()

        matching = nx.algorithms.matching.max_weight_matching(graph)
        matching = map(lambda x: x if x[0].startswith('x') else (x[1], x[0]), matching)

        matched_indices = [(self._convert(vertex_a), self._convert(vertex_b)) for vertex_a, vertex_b in matching]
        matched_indices_arr = np.array(matched_indices).transpose()

        final_score = scores[matched_indices_arr[0], matched_indices_arr[1]].sum() / max_scores
        matched = []
        for vertex_a, vertex_b in matched_indices:
            if isinstance(flat_from_module[vertex_b], list) and isinstance(flat_to_module[vertex_a], list):
                matched.append(self._match_models(flat_from_module[vertex_b], flat_to_module[vertex_a])[1])
            else:
                matched.append((flat_from_module[vertex_b], flat_to_module[vertex_a]))

        return final_score, matched, matched_indices

    @staticmethod
    def _convert(vertex: str) -> int:
        return int(vertex[1:]) if vertex.startswith('x') else int(vertex)



import pkgutil

__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    __all__.append(module_name)
    _module = loader.find_module(module_name).load_module(module_name)
    globals()[module_name] = _module

from .iat import IAT


from __future__ import annotations

import random
from collections import Counter
from typing import List, Tuple, Union, Dict

import torch
import torch.nn as nn
from graphviz import Digraph
from torch.autograd import Variable

from iatransfer.toolkit.base_standardization import Standardization


class GraphStandardization(Standardization):
    """Graph standardization algorithm for IAT.
    """

    def standardize(self, module: nn.Module, *args, **kwargs) \
            -> List[Union[nn.Module, List[nn.Module]]]:
        return self.get_blocks(module)

    def make_dot(self, var, params=None):

        def resize_graph(dot, size_per_element=0.15, min_size=12):
            num_rows = len(dot.body)
            content_size = num_rows * size_per_element
            size = max(min_size, content_size)
            size_str = str(size) + "," + str(size)
            dot.graph_attr.update(size=size_str)
            return size

        if params is not None:
            assert all(isinstance(p, Variable) for p in params.values())
            param_map = {id(v): k for k, v in params.items()}

        node_attr = dict(
            style="filled",
            shape="box",
            align="left",
            fontsize="12",
            ranksep="0.1",
            height="0.2",
        )

        idx = random.randint(1, 100000)
        dot = Digraph(
            name=f"cluster_{idx}", node_attr=node_attr, graph_attr=dict(size="12,12")
        )
        seen = set()

        def size_to_str(size):
            return "(" + (", ").join(["%d" % v for v in size]) + ")"

        mod_op = ["AddBackward0", "MulBackward0", "CatBackward"]

        lmap, emap, hmap = {}, {}, {}
        elist, clist = [], []

        def add_nodes(var, root=None, c=0, depth=0, branch=0, global_i=0):
            if var in seen:
                return None

            depth += 1
            global_i += 1

            if c not in lmap:
                lmap[c], emap[c] = [], []

            # FIXME: move to function
            if hasattr(var, "variable"):
                u = var.variable
                name = param_map[id(u)] if params is not None else ""
                node_name = "%s\n %s" % (name, size_to_str(u.size()))
                lmap[c].append(
                    {
                        "branch": branch,
                        "depth": depth,
                        "global_i": global_i,
                        "name": name,
                        "size": u.size(),
                        "type": "param",
                        "id": str(id(var)),
                    }
                )
                hmap[name] = str(id(var))
                dot.node(
                    str(id(var)),
                    f"c={c} branch={branch} depth={depth}\n" + node_name,
                    fillcolor="lightblue",
                )
            else:
                node_name = str(type(var).__name__)
                if node_name in mod_op:
                    depth = 0
                    prev_c = c
                    c = str(id(var))
                    clist.append((c, prev_c))
                    emap[c], lmap[c] = [], []
                    dot.node(str(id(var)), node_name + f" [{c}]", fillcolor="green")
                else:
                    lmap[c].append(
                        {
                            "branch": branch,
                            "depth": depth,
                            "global_i": global_i,
                            "name": node_name,
                            "type": "func",
                            "id": str(id(var)),
                        }
                    )
                    dot.node(str(id(var)), node_name + f" [{str(id(var))}]")
            if root:
                emap[c].append((str(id(var)), root))
            seen.add(var)

            if hasattr(var, "next_functions"):
                for _branch, u in enumerate(var.next_functions):
                    if node_name in mod_op:
                        branch = _branch
                    if u[0] is not None:
                        dot.edge(str(id(u[0])), str(id(var)), color="blue")
                        elist.append((str(id(u[0])), str(id(var))))
                        add_nodes(
                            u[0],
                            root=str(id(var)),
                            c=c,
                            depth=depth,
                            branch=branch,
                            global_i=global_i,
                        )

        if isinstance(var, tuple):
            for v in var:
                add_nodes(v.grad_fn)
        else:
            add_nodes(var.grad_fn)

        for c, edges in emap.items():
            with dot.subgraph(name=f"cluster_{idx}_{c}") as _c:
                _c.attr(color="blue")
                _c.attr(style="filled", color="lightgrey")
                _c.node_attr["style"] = "filled"
                _c.edges(edges)
                _c.attr(label=f"cluster {c}")

        return dot, lmap, hmap, clist, elist

    def get_graph(self, model: nn.Module):
        # FIXME: warning about 'torchviz'
        try:
            x = torch.randn(32, 3, 32, 32)
            g, lmap, hmap, clist, elist = self.make_dot(model(x), params=dict(model.named_parameters()))
        except:
            # FIXME: universal head? (what happens if MNIST?)
            print("ERROR: trying one channel")
            x = torch.randn(32, 1, 31, 31)
            g, lmap, hmap, clist, elist = self.make_dot(model(x), params=dict(model.named_parameters()))
        return g, lmap, hmap, clist, elist

    def get_layers(self, model: nn.Module) -> Dict[str, nn.Module]:
        layers_dict = {}

        def dfs(model: nn.Module, name_prefix: List[str]):
            for child_name, child in model.named_children():
                dfs(child, name_prefix + [child_name])
            layers_dict[".".join(name_prefix)] = model

        dfs(model, [])
        return layers_dict

    def sort_graph(self, graph: 'Graph', clusters_edges: List[Tuple[str, str]],
                   nodes_edges: List[Tuple[str, str]]) -> 'Graph':

        def topological_sort(edges: List[Tuple[str, str]]) -> Dict[str, int]:
            def dfs(a: 'Node'):
                nonlocal postorder
                a['vis'] = True
                for bid in a['e']:
                    b = nodes[bid]
                    if not b['vis']:
                        dfs(b)
                a['postorder'] = postorder
                postorder += 1

            nodes = set([e[0] for e in edges]) | set([e[1] for e in edges])
            nodes = dict([
                (node, {'id': node, 'e': [], 'postorder': -1, 'vis': False}) for node in nodes
            ])
            for e in edges:
                nodes[e[0]]['e'].append(e[1])
            postorder = 0
            for node in nodes.values():
                if node['postorder'] == -1:
                    dfs(node)
            ordered_nodes = sorted(list(nodes.values()), key=lambda node: -node['postorder'])
            return dict([(node['id'], i) for i, node in enumerate(ordered_nodes)])

        def count_ancestors(edges: List[Tuple[str, str]]) -> Dict[str, int]:
            return

        def sort_cluster(cluster: List['Node']) -> List['Node']:
            def cmp(node: 'Node') -> int:
                if node['type'] == 'param':
                    a = node_child[node['id']]
                    while parents_count[a] == 1:
                        a = node_child[a]
                    return nodes_order[a]
                else:
                    return nodes_order[node["id"]]

            return sorted(cluster, key=cmp)

        clusters_order = topological_sort(clusters_edges)
        node_child = dict(nodes_edges)
        nodes_order = topological_sort(nodes_edges)
        parents_count = Counter([e[1] for e in nodes_edges])

        new_graph = {}
        for cid in clusters_order.keys():
            cluster = graph[cid]
            cluster = sort_cluster(cluster)
            new_graph[cid] = cluster
        return new_graph

    def get_blocks(self, model: nn.Module) -> List[Union[nn.Module, List[nn.Module]]]:
        _, graph, _, clusters_edges, nodes_edges = self.get_graph(model)
        graph = self.sort_graph(graph, clusters_edges, nodes_edges)
        layers_dict = self.get_layers(model)
        layers = []
        for block in graph.values():
            layers_in_block = []
            for layer in block:
                if layer['name'].endswith('.weight'):
                    name = layer['name'].replace('.weight', '')
                    layers_in_block.append(layers_dict[name])
            layers.append(layers_in_block)
        return layers



import distutils.dir_util
import glob
import os
import sys
import tarfile
from pathlib import Path
from typing import List

from setuptools import setup, find_packages

from iatransfer.utils.dot_dict import DotDict
from iatransfer.utils.file_utils import read_json, read_contents

PROPERTIES = 'setup_configuration.json'
DISTRIBUTIONS = 'DISTRIBUTIONS'


def forbid_publish() -> None:
    argv = sys.argv
    blacklist = ['register', 'upload']

    for command in blacklist:
        if command in argv:
            values = {'command': command}
            print('Command "%(command)s" has been blacklisted, exiting...' %
                  values)
            sys.exit(2)


def mkdir(path: str) -> None:
    Path(path).mkdir(parents=True, exist_ok=True)


def cleanup(path: str) -> None:
    try:
        distutils.dir_util.remove_tree(path)
    except FileNotFoundError:
        pass


def requirements(filename: str) -> List[str]:
    return [i.strip() for i in open(filename).readlines()]


def do_setup(configuration: DotDict) -> None:
    print(configuration.DATA_FILES)
    setup(
        name=configuration.NAME,
        version=configuration.VERSION,
        author=configuration.AUTHOR,
        author_email=configuration.EMAIL,
        url=configuration.URL,
        description=configuration.DESCRIPTION,
        long_description=read_contents(configuration.README_PATH),
        long_description_content_type=configuration.LONG_DESCRIPTION_CONTENT_TYPE,
        license=configuration.LICENSE,
        packages=find_packages(include=configuration.PACKAGES),
        entry_points=configuration.ENTRYPOINTS,
        scripts=[x for string in configuration.SCRIPTS for x in glob.glob(string)],
        classifiers=configuration.CLASSIFIERS,
        keywords=configuration.KEYWORDS,
        install_requires=requirements(configuration.REQUIREMENTS_PATH),
        zip_safe=False,
        data_files=configuration.DATA_FILES
    )


def read_configuration(name: str) -> DotDict:
    properties = read_json(PROPERTIES)
    properties = {**properties, **properties[DISTRIBUTIONS][name]}
    del properties[DISTRIBUTIONS]
    return DotDict(properties)


def setup_configuration(name: str) -> None:
    # TODO remove once we can publish
    forbid_publish()

    configuration = read_configuration(name)
    cleanup(configuration.BUILD_DIR_PATH)

    do_setup(configuration)


def replace_setuppy(configuration: DotDict) -> None:
    mkdir(configuration.DIST_DIR_PATH)

    name = glob.glob(f'{configuration.DIST_DIR_PATH}/iatransfer_research-*.tar.gz')[0]
    dest_name = f'{configuration.DIST_DIR_PATH}/iatransfer_research-tmp.tar.gz'

    tar_in = tarfile.open(name, 'r:gz')
    tar_out = tarfile.open(dest_name, 'w:gz')

    for item in tar_in.getmembers():
        name_split = item.name.split('/')
        if len(name_split) > 1:
            if name_split[1] != 'setup.py':
                tar_out.addfile(item, tar_in.extractfile(item))
            if name_split[1] == 'setup_research.py':
                item.name = f'{name_split[0]}/setup.py'
                tar_out.addfile(item, tar_in.extractfile(item))

    tar_in.close()
    tar_out.close()

    os.rename(dest_name, name)



from typing import Tuple, List, Any, Optional

import numpy as np
import torch.nn as nn

from iatransfer.toolkit.base_matching import Matching


class DPFlatMatching(Matching):
    """Dynamic programming flat matching algorithm for IAT.
    """

    def match(self, from_module: List[nn.Module],
              to_module: List[nn.Module], *args, **kwargs) \
            -> List[Tuple[nn.Module, nn.Module]]:
        _, matched, _ = self._match_models(from_module, to_module)
        return matched

    def _compute_score(self, from_module: nn.Module, to_module: nn.Module) -> float:
        def are_all_of_this_class(layers: List[nn.Module], clazz: Any) -> bool:
            return all([isinstance(layer, clazz) for layer in layers])

        score = 0
        layers = [from_module, to_module]
        classes = [
            nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
            nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
            nn.Linear
        ]
        for clazz in classes:
            if are_all_of_this_class(layers, clazz):
                score = 1
                for x, y in zip(from_module.weight.shape, to_module.weight.shape):
                    score *= min(x / y, y / x)
                break

        return score

    def _penalty(self, x):
        return -(x + 1) / 2 if x >= 0 else 0

    def _match_models(self, from_module: List[nn.Module], to_module: List[nn.Module]) -> List[
        Tuple[Optional[nn.Module]]]:
        n = len(to_module)
        m = len(from_module)
        dp = np.zeros((n + 1, m + 1))
        transition = np.zeros((n + 1, m + 1, 2), dtype=int)
        for i in range(1, n + 1):
            for j in range(1, m + 1):
                score = self._compute_score(from_module[j - 1], to_module[i - 1])
                if dp[i, j - 1] > dp[i - 1, j]:
                    dp[i, j] = dp[i, j - 1]
                    transition[i, j] = [i, j - 1]
                else:
                    dp[i, j] = dp[i - 1, j]
                    transition[i, j] = [i - 1, j]
                for k in range(0, m + 1):  # row
                    value = dp[i - 1, k] + score + self._penalty(k - j)
                    if value > dp[i, j]:
                        dp[i, j] = value
                        transition[i, j] = [i - 1, k]

        matched, matched_indices = [], []
        i, j = n, m
        while i > 0 and j > 0:
            if transition[i, j, 0] == i:
                matched.append((from_module[j - 1], None))
                matched_indices.append((j - 1, None))
            else:
                if dp[tuple(transition[i, j])] != dp[i, j]:
                    matched.append((from_module[j - 1], to_module[i - 1]))
                    matched_indices.append((j - 1, i - 1))
                else:
                    matched.append((None, to_module[i - 1]))
                    matched_indices.append((None, i - 1))
            i, j = transition[i, j]
        while i > 0:
            matched.append((None, to_module[i - 1]))
            matched_indices.append((None, i - 1))
            i -= 1
        while j > 0:
            matched.append((to_module[j - 1], None))
            matched_indices.append((j - 1, None))
            j -= 1

        matched.reverse()
        matched_indices.reverse()
        return dp[n][m], matched, matched_indices



from typing import Dict, List

from torch import nn

from iatransfer.toolkit.tli import get_graph


def get_model_graph_and_ids_mapping(model: nn.Module) -> ['Graph', Dict[nn.Module, int]]:
    graph = get_graph(model)
    names_to_layers_mapping = {}
    def dfs(model: nn.Module, name_prefix: List[str]):
        for child_name, child in model.named_children():
            dfs(child, name_prefix + [child_name])
        names_to_layers_mapping[".".join(name_prefix)] = model
    dfs(model, [])
    
    layers_to_ids_mapping = {}
    for node in graph.nodes.values():
        if node.name.endswith(".weight"):
            layer = names_to_layers_mapping[node.name.replace(".weight", "")]
            layers_to_ids_mapping[layer] = (node.idx, layers_to_ids_mapping.get(layer, (None, None))[1])
        elif node.name.endswith(".bias"):
            layer = names_to_layers_mapping[node.name.replace(".bias", "")]
            layers_to_ids_mapping[layer] = (layers_to_ids_mapping.get(layer, (None, None))[0], node.idx)
    
    return graph, layers_to_ids_mapping


from typing import List, Union

import torch.nn as nn

from iatransfer.toolkit.base_standardization import Standardization


class FlattenStandardization(Standardization):
    """Basic flatten standardization algorithm for IAT.
    """

    def standardize(self, module: nn.Module, *args, **kwargs) \
            -> List[Union[nn.Module, List[nn.Module]]]:
        classes = [
            nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
            nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
            nn.Linear
        ]
        if any([isinstance(module, clazz) for clazz in classes]):
            return [module]

        layers = []
        for child in module.children():
            layers += self.standardize(child)
        if len(layers) == 0:
            layers = [module]
        return layers



from torch import nn

from iatransfer.toolkit.base_score import Score


class ShapeScore(Score):
    """Shape based score algorithm for IAT.
    """

    def score(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> float:
        score = 1
        for x, y in zip(from_module.weight.shape, to_module.weight.shape):
            score *= min(x / y, y / x)
        return score



from typing import Tuple, List, Union, Dict

import torch.nn as nn

from iatransfer.toolkit.base_matching import Matching
from iatransfer.toolkit.tli import transfer, get_graph, Graph


class GraphNewMatching(Matching):
    """PyTorch execution graph analysis matching algorithm for IAT.
    """

    def match(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> List[Union[Tuple[nn.Module, nn.Module], List[Tuple[nn.Module, nn.Module]]]]:
        from_module = kwargs['context']['from_module']
        to_module = kwargs['context']['to_module']
        _, remap, teacher_graph, student_graph = transfer(from_module, to_module)
        teacher_ids_to_layers_mapping = self.get_ids_to_layers_mapping(from_module, teacher_graph)
        student_ids_to_layers_mapping = self.get_ids_to_layers_mapping(to_module, student_graph)
        matching = []
        total, matched = 0, 0
        classes = [
            nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
            nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
            nn.Linear
        ]
        for student_idx, teacher_idx in remap.items():
            total+=1
            try:
                teacher_layer = teacher_ids_to_layers_mapping[teacher_idx]
                student_layer = student_ids_to_layers_mapping[student_idx]
                for clazz in classes:
                    if isinstance(teacher_layer, clazz) and isinstance(student_layer, clazz):
                        matching.append((teacher_layer, student_layer)) 
                        matched+=1
                        break
            except:
                pass
        print("Matched: ", matched/total)
        return matching
    
    def sim(self, from_module: nn.Module, to_module: nn.Module) -> float:
        return transfer(from_module, to_module)[0]

    def get_ids_to_layers_mapping(self, model: nn.Module, graph: Graph) -> Dict[int, nn.Module]:
        graph = get_graph(model)
        names_to_layers_mapping = {}
        def dfs(model: nn.Module, name_prefix: List[str]):
            for child_name, child in model.named_children():
                dfs(child, name_prefix + [child_name])
            names_to_layers_mapping[".".join(name_prefix)] = model
        dfs(model, [])
        
        ids_to_layers_mapping = {}
        for node in graph.nodes.values():
            if node.name.endswith(".weight"):
                layer = names_to_layers_mapping[node.name.replace(".weight", "")]
                ids_to_layers_mapping[node.idx] = layer
                
        return ids_to_layers_mapping



from typing import Any, Dict, Tuple, Union, List, Set

import torch.nn as nn
from inflection import camelize

from iatransfer.toolkit.base_matching import Matching
from iatransfer.toolkit.base_score import Score
from iatransfer.toolkit.base_standardization import Standardization
from iatransfer.toolkit.base_transfer import Transfer
from iatransfer.toolkit.transfer.transfer_stats import TransferStats
from iatransfer.utils.dot_dict import DotDict
from iatransfer.utils.flatten import flatten_modules
from iatransfer.utils.subclass_utils import get_subclasses


class IAT:
    """Represents the inter-architecture transfer algorithm.
    """
    standardization: Standardization = None
    matching: Matching = None
    transfer: Transfer = None
    score: Score = None

    def run(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> TransferStats:
        """Executes the entire IAT algorithm.
        """
        context = {'from_module': from_module, 'to_module': to_module}

        flat_from = set(flatten_modules(from_module))
        flat_to = set(flatten_modules(to_module))

        from_paths, to_paths = self.standardization(from_module), self.standardization(to_module)
        self.score.precompute_scores(from_module, to_module)
        self.matching.set_score(self.score)
        matched_tensors = self.matching(from_paths, to_paths, context=context)

        self.transfer(matched_tensors, context=context)

        all_from = len(flat_from)
        all_to = len(flat_to)
        self._flat_remove(matched_tensors, flat_from, flat_to)

        return TransferStats(all_from=all_from,
                             all_to=all_to,
                             left_from=len(flat_from),
                             left_to=len(flat_to),
                             matched_from=all_from - len(flat_from),
                             matched_to=all_to - len(flat_to))

    def _flat_remove(self, matched: List, flat_from: Set[nn.Module], flat_to: Set[nn.Module]):
        for x in matched:
            if isinstance(x, list):
                self._flat_remove(x, flat_from, flat_to)
            else:
                tensor_from, tensor_to = x
                if tensor_to and tensor_from:
                    if tensor_from in flat_from:
                        flat_from.remove(tensor_from)
                    if tensor_to in flat_to:
                        flat_to.remove(tensor_to)

    def sim(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> float:
        """Returns similarity score between two models.
        """
        from_paths, to_paths = self.standardization(from_module), self.standardization(to_module)
        self.score.precompute_scores(from_module, to_module)
        self.matching.set_score(self.score)
        return self.matching.sim(from_paths, to_paths)

    def __call__(self, from_module: nn.Module, to_module: nn.Module, *args, **kwargs) \
            -> TransferStats:
        """Alias for 'run'.
        """
        return self.run(from_module, to_module, *args, **kwargs)

    def __init__(self,
                 standardization: Union[Standardization, str, Tuple[Standardization, Dict], Tuple[str, Dict]]
                 = 'blocks',
                 matching: Union[Matching, str, Tuple[Matching, Dict], Tuple[str, Dict]] = 'dp',
                 transfer: Union[Transfer, str, Tuple[Transfer, Dict], Tuple[str, Dict]] = 'clip',
                 score: Union[Score, str, Tuple[Score, Dict], Tuple[str, Dict]] = 'ShapeScore',
                 *args,
                 **kwargs):
        ctx = DotDict()

        ctx._standardization_classes = get_subclasses(Standardization)
        ctx._matching_classes = get_subclasses(Matching)
        ctx._transfer_classes = get_subclasses(Transfer)
        ctx._score_classes = get_subclasses(Score)

        self._try_setting(ctx, 'standardization', standardization, Standardization)
        self._try_setting(ctx, 'matching', matching, Matching)
        self._try_setting(ctx, 'transfer', transfer, Transfer)
        self._try_setting(ctx, 'score', score, Score)

    @staticmethod
    def make(standardization: Union[Standardization, str, Tuple[Standardization, Dict], Tuple[str, Dict]]
             = 'blocks',
             matching: Union[Matching, str, Tuple[Matching, Dict], Tuple[str, Dict]] = 'dp',
             transfer: Union[Transfer, str, Tuple[Transfer, Dict], Tuple[str, Dict]] = 'clip',
             score: Union[Score, str, Tuple[Score, Dict], Tuple[str, Dict]] = 'ShapeScore',
             *args,
             **kwargs):
        """Method factory.
        """
        return IAT(standardization, matching, transfer, score, *args, **kwargs)

    def _try_setting(self, ctx: Dict, key: str, value: Any, clazz: Any) -> None:
        kwargs = {}
        if isinstance(value, tuple):
            value, kwargs = value[0], value[1]
        if isinstance(value, str):
            setattr(self, key, self._find_subclass(ctx, key, value)(**kwargs))
        elif isinstance(value, clazz):
            setattr(self, key, value)
        else:
            raise TypeError()

    def _find_subclass(self, ctx: Dict, key: str, value: str) -> Any:
        classes = getattr(ctx, f'_{key}_classes')
        trials = [lambda x: x[1],
                  lambda x: camelize(f'{x[1]}_{x[0]}'),
                  lambda x: camelize(f'{camelize(x[1])}{camelize(x[0])}'),
                  lambda x: camelize(f'{x[1].upper()}{camelize(x[0])}'),
                  lambda x: f'{x[1].split("_")[0].upper()}{camelize("_".join(x[1].split("_")[1:]))}{camelize(x[0])}'
                  ]
        for trial in trials:
            res = classes.get(trial((key, value)))
            if res is not None:
                return res
        raise ValueError()



from typing import List, Tuple, Union

import torch.nn as nn

from iatransfer.toolkit.base_standardization import Standardization


class BlocksStandardization(Standardization):
    """Block standardization algorithm for IAT.
    """

    def standardize(self, module: nn.Module, *args, **kwargs) \
            -> List[Union[nn.Module, List[nn.Module]]]:
        return self.flatten_with_blocks(module, 0)[2]

    def flatten_with_blocks(self, module: nn.Module, level: int = 0) -> Tuple[int, int, List[nn.Module]]:
        depth, dimensional_layers_num, layers = 0, 0, []
        classes = [
            nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
            nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
            nn.Linear
        ]
        dimensional_layers = [
            nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
            nn.Linear
        ]
        if any([isinstance(module, clazz) for clazz in classes]):
            layers = module
            if any([isinstance(module, clazz) for clazz in dimensional_layers]):
                dimensional_layers_num += 1
        else:
            blocks = 0
            single_layers = 0
            children = list(module.children())
            if len(children) == 1:
                next_level = level
            else:
                next_level = level + 1
            for child in module.children():
                child_depth, child_dimensional_layers_num, child_layers = self.flatten_with_blocks(child, next_level)
                dimensional_layers_num += child_dimensional_layers_num
                if child_depth > 1 or (isinstance(child_layers, list) and child_dimensional_layers_num < 2):
                    for child_layer in child_layers:
                        if isinstance(child_layer, list):
                            blocks += 1
                        else:
                            single_layers += 1
                        layers += [child_layer]
                else:
                    if isinstance(child_layers, list):
                        blocks += 1
                    else:
                        single_layers += 1
                    layers += [child_layers]

            if single_layers > blocks and level > 0:
                new_layers = []
                for child_layers in layers:
                    if isinstance(child_layers, list):
                        new_layers += child_layers
                    else:
                        new_layers += [child_layers]
                layers = new_layers
                depth = 1
            else:
                depth = 2

            if len(layers) == 1:
                layers = layers[0]
                depth = 1
            if isinstance(layers, list) and len(layers) == 0:
                depth = 0
                layers = module

        if level == 0 and isinstance(layers, list) is False:
            layers = [layers]

        return depth, dimensional_layers_num, layers



from abc import ABC, abstractmethod
from typing import List, Tuple, Union

import torch.nn as nn

from iatransfer.toolkit.base_score import Score


class Matching(ABC):
    """Base-class for any matching algorithm detectable from IAT.
    """

    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs

    def sim(self, from_module: List[Union[nn.Module, List[nn.Module]]],
            to_module: List[Union[nn.Module, List[nn.Module]]], *args, **kwargs) -> float:
        """ Calculates the similarity between two standardized models (lists of blocks).

        :return: Float in range [0; 1] representing similarity score.
        """
        raise NotImplementedError()

    @abstractmethod
    def match(self, from_module: List[Union[nn.Module, List[nn.Module]]],
              to_module: List[Union[nn.Module, List[nn.Module]]], *args, **kwargs) \
            -> List[Tuple[nn.Module, nn.Module]]:
        """ Matches two standardized models (lists of blocks).

        :return: The list of tuples of matched layers.
        """
        pass

    def __call__(self, from_module: List[Union[nn.Module, List[nn.Module]]],
                 to_module: List[Union[nn.Module, List[nn.Module]]], *args, **kwargs) \
            -> List[Tuple[nn.Module, nn.Module]]:
        """ Alias for 'match'.
        """
        return self.match(from_module, to_module, *args, **kwargs)

    def set_score(self, score: Score) -> None:
        self.score = score


from abc import ABC, abstractmethod
from typing import List, Union

import torch.nn as nn


class Standardization(ABC):
    """Base-class for any standardization algorithm detectable from IAT.
    """

    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs

    @abstractmethod
    def standardize(self, module: nn.Module, *args, **kwargs) \
            -> List[Union[nn.Module, List[nn.Module]]]:
        """Standardizes the model to a list of blocks.
        """
        pass

    def __call__(self, module: nn.Module, *args, **kwargs) \
            -> List[Union[nn.Module, List[nn.Module]]]:
        """Alias for standardize.
        """
        return self.standardize(module, *args, **kwargs)



from typing import List

from torch import nn


def flatten_modules(module: nn.Module) -> List[nn.Module]:
    classes = [
        nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d,
        nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
        nn.Linear
    ]
    if any([isinstance(module, clazz) for clazz in classes]):
        return [module]

    layers = []
    for child in module.children():
        layers.extend(flatten_modules(child))
    if len(layers) == 0:
        layers.append(module)
    return layers



from typing import Dict


def get_subclasses(clazz: type) -> Dict:
    subclasses = {}
    classes = [clazz]
    while len(classes) > 0:
        clazz = classes.pop()
        for c in clazz.__subclasses__():
            subclasses[c.__name__] = c
            classes.append(c)
    return subclasses
